;; Generated package description from modus-themes.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "modus-themes" "5.3.0.0.20260831.40" "Elegant, highly legible and customizable themes" '((emacs "28.1")) :commit "24e4b4b9e47a382cba6dfb3dc41d68a8ba7912d2" :authors '(("Protesilaos" . "info@protesilaos.com")) :maintainer '("Protesilaos" . "info@protesilaos.com") :keywords '("faces" "theme" "accessibility") :url "https://github.com/protesilaos/modus-themes")
